#!/usr/bin/env python3
import numpy as np
import pickle
from pathlib import Path
from .flash_entropy_search_core import FlashEntropySearchCore
from .flash_entropy_search_core_low_memory import FlashEntropySearchCoreLowMemory
from .flash_entropy_search_core_medium_memory import FlashEntropySearchCoreMediumMemory
from ..spectra import clean_spectrum


class FlashEntropySearch:
    def __init__(
        self,
        max_ms2_tolerance_in_da=0.024,
        mz_index_step=0.0001,
        low_memory=False,
        path_data=None,
        intensity_weight="entropy",
        **kwargs,
    ):
        """
        Initialize the EntropySearch class.
        
        :param max_ms2_tolerance_in_da:  The maximum MS2 tolerance in Da.
        :param mz_index_step:   The step size for the m/z index.
        :param low_memory:  The memory usage mode, can be 0, 1, or 2. 0 means normal mode, 1 means low memory mode, and 2 means medium memory mode.
        :param path_data:   The path to save the index data.
        :param intensity_weight:    The weight for the intensity in the entropy calculation, can be "entropy" or None. Default is "entropy".
            - None: The intensity will not be weighted, then the unweighted similarity will be calculated.
            - "entropy": The intensity will be weighted by the entropy, then the entropy similarity will be calculated.
        :param kwargs:  Those parameters will be ignored.
        """
        self.precursor_mz_array = np.zeros(0, dtype=np.float32)
        self.low_memory = low_memory
        if low_memory == 1:
            self.entropy_search = FlashEntropySearchCoreLowMemory(
                path_data=path_data, max_ms2_tolerance_in_da=max_ms2_tolerance_in_da, mz_index_step=mz_index_step, intensity_weight=intensity_weight
            )
        elif low_memory == 2:
            self.entropy_search = FlashEntropySearchCoreMediumMemory(
                path_data=path_data, max_ms2_tolerance_in_da=max_ms2_tolerance_in_da, mz_index_step=mz_index_step, intensity_weight=intensity_weight
            )
        else:
            self.entropy_search = FlashEntropySearchCore(
                path_data=path_data, max_ms2_tolerance_in_da=max_ms2_tolerance_in_da, mz_index_step=mz_index_step, intensity_weight=intensity_weight
            )

    def identity_search(self, precursor_mz, peaks, ms1_tolerance_in_da, ms2_tolerance_in_da, target="cpu", output_matched_peak_number=False, **kwargs):
        """
        Run the identity search, the query spectrum should be preprocessed by `clean_spectrum()` function before calling this function.

        For super large spectral library, directly identity search is not recommended. To do the identity search on super large spectral library,
        divide the spectral library into several parts, build the index for each part, and then do the identity search on each part will be much faster.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms1_tolerance_in_da:  The MS1 tolerance in Da.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :param target:  The target device for the search, can be "cpu" or "gpu".
        :param output_matched_peak_number:  If True, the number of matched peaks will be returned with the entropy similarity score.

        :return:    The entropy similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
                    If `output_matched_peak_number` is True, the number of matched peaks will be returned with the entropy similarity score, i.e. the return
                    will be a tuple of two numpy arrays, the first one is the entropy similarity score, and the second one is the number of matched peaks.
        """
        precursor_mz_min = precursor_mz - ms1_tolerance_in_da
        precursor_mz_max = precursor_mz + ms1_tolerance_in_da
        spectra_idx_min = np.searchsorted(self.precursor_mz_array, precursor_mz_min, side="left")
        spectra_idx_max = np.searchsorted(self.precursor_mz_array, precursor_mz_max, side="right")
        if spectra_idx_min >= spectra_idx_max:
            if output_matched_peak_number:
                return np.zeros(self.entropy_search.total_spectra_num, dtype=np.float32), np.zeros(self.entropy_search.total_spectra_num, dtype=np.uint16)
            else:
                return np.zeros(self.entropy_search.total_spectra_num, dtype=np.float32)
        else:
            return self.entropy_search.search(
                method="open",
                target=target,
                peaks=peaks,
                ms2_tolerance_in_da=ms2_tolerance_in_da,
                search_type=1,
                search_spectra_idx_min=spectra_idx_min,
                search_spectra_idx_max=spectra_idx_max,
                output_matched_peak_number=output_matched_peak_number,
            )

    def open_search(self, peaks, ms2_tolerance_in_da, target="cpu", output_matched_peak_number=False, **kwargs):
        """
        Run the open search, the query spectrum should be preprocessed by `clean_spectrum()` function before calling this function.

        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :param target:  The target device for the search, can be "cpu" or "gpu".
        :param output_matched_peak_number:  If True, the number of matched peaks will be returned with the entropy similarity score.

        :return:    The entropy similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
                    If `output_matched_peak_number` is True, the number of matched peaks will be returned with the entropy similarity score, i.e. the return
                    will be a tuple of two numpy arrays, the first one is the entropy similarity score, and the second one is the number of matched peaks.
        """
        return self.entropy_search.search(
            method="open",
            target=target,
            peaks=peaks,
            ms2_tolerance_in_da=ms2_tolerance_in_da,
            search_type=0,
            output_matched_peak_number=output_matched_peak_number,
        )

    def neutral_loss_search(self, precursor_mz, peaks, ms2_tolerance_in_da, target="cpu", output_matched_peak_number=False, **kwargs):
        """
        Run the neutral loss search, the query spectrum should be preprocessed by `clean_spectrum()` function before calling this function.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :param target:  The target device for the search, can be "cpu" or "gpu".
        :param output_matched_peak_number:  If True, the number of matched peaks will be returned with the entropy similarity score.

        :return:    The entropy similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
                    If `output_matched_peak_number` is True, the number of matched peaks will be returned with the entropy similarity score, i.e. the return
                    will be a tuple of two numpy arrays, the first one is the entropy similarity score, and the second one is the number of matched peaks.
        """
        return self.entropy_search.search(
            method="neutral_loss",
            target=target,
            precursor_mz=precursor_mz,
            peaks=peaks,
            ms2_tolerance_in_da=ms2_tolerance_in_da,
            search_type=0,
            output_matched_peak_number=output_matched_peak_number,
        )

    def hybrid_search(self, precursor_mz, peaks, ms2_tolerance_in_da, target="cpu", **kwargs):
        """
        Run the hybrid search, the query spectrum should be preprocessed by `clean_spectrum()` function before calling this function.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be the output of `clean_spectrum()` function.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da.
        :param target:  The target device for the search, can be "cpu" or "gpu".

        :return:    The entropy similarity score for each spectrum in the library, a numpy array with shape (N,), N is the number of spectra in the library.
        """
        return self.entropy_search.search_hybrid(target=target, precursor_mz=precursor_mz, peaks=peaks, ms2_tolerance_in_da=ms2_tolerance_in_da)

    def clean_spectrum_for_search(
        self, precursor_mz, peaks, precursor_ions_removal_da: float = 1.6, noise_threshold=0.01, min_ms2_difference_in_da: float = 0.05, max_peak_num: int = 0
    ):
        """
        Clean the MS/MS spectrum, need to be called before any search.

        :param precursor_mz:    The precursor m/z of the spectrum.
        :param peaks:           The peaks of the spectrum, should be a list or numpy array with shape (N, 2), N is the number of peaks. The format of the peaks is [[mz1, intensity1], [mz2, intensity2], ...].
        :param precursor_ions_removal_da:   The ions with m/z larger than precursor_mz - precursor_ions_removal_da will be removed.
                                            Default is 1.6. Set to None to not remove any ions.
        :param noise_threshold: The intensity threshold for removing the noise peaks. The peaks with intensity smaller than noise_threshold * max(intensity)
                                will be removed. Default is 0.01.
        :param min_ms2_difference_in_da:    The minimum difference between two peaks in the MS/MS spectrum. Default is 0.05.
        :param max_peak_num:    The maximum number of peaks in the MS/MS spectrum. Default is 0, which means no limit.
        """
        if precursor_ions_removal_da is not None:
            max_mz = precursor_mz - precursor_ions_removal_da
        else:
            max_mz = None
        return clean_spectrum(
            peaks=peaks,
            min_mz=-1,
            max_mz=max_mz,
            noise_threshold=noise_threshold,
            min_ms2_difference_in_da=min_ms2_difference_in_da,
            max_peak_num=max_peak_num,
            normalize_intensity=True,
        )

    def search(
        self,
        precursor_mz,
        peaks,
        ms1_tolerance_in_da=0.01,
        ms2_tolerance_in_da=0.02,
        method="all",  # "identity", "open", "neutral_loss", "hybrid", "all", or list of the above
        target="cpu",
        precursor_ions_removal_da: float = 1.6,
        noise_threshold=0.01,
        min_ms2_difference_in_da: float = 0.05,
        max_peak_num: int = None,
    ):
        """
        Run the Flash entropy search for the query spectrum.

        :param precursor_mz:    The precursor m/z of the query spectrum.
        :param peaks:           The peaks of the query spectrum, should be a list or numpy array with shape (N, 2), N is the number of peaks. The format of the peaks is [[mz1, intensity1], [mz2, intensity2], ...].
        :param ms1_tolerance_in_da:  The MS1 tolerance in Da. Default is 0.01.
        :param ms2_tolerance_in_da:  The MS2 tolerance in Da. Default is 0.02.
        :param method:  The search method, can be "identity", "open", "neutral_loss", "hybrid", "all", or list of the above.
        :param target:  The target device for the search, can be "cpu" or "gpu".
        :param precursor_ions_removal_da:   The ions with m/z larger than precursor_mz - precursor_ions_removal_da will be removed.
                                            Default is 1.6.
        :param noise_threshold: The intensity threshold for removing the noise peaks. The peaks with intensity smaller than noise_threshold * max(intensity)
                                will be removed. Default is 0.01.
        :param min_ms2_difference_in_da:    The minimum difference between two peaks in the MS/MS spectrum. Default is 0.05.
        :param max_peak_num:    The maximum number of peaks in the MS/MS spectrum. Default is None, which means no limit.

        :return:    A dictionary with the search results. The keys are "identity_search", "open_search", "neutral_loss_search", "hybrid_search", and the values are the search results for each method.
        """
        if precursor_ions_removal_da is not None:
            max_mz = precursor_mz - precursor_ions_removal_da
        else:
            max_mz = -1
        peaks = clean_spectrum(
            peaks=peaks,
            min_mz=0,
            max_mz=max_mz,
            noise_threshold=noise_threshold,
            min_ms2_difference_in_da=min_ms2_difference_in_da,
            max_peak_num=max_peak_num,
            normalize_intensity=True,
        )
        if method == "all":
            method = {"identity", "open", "neutral_loss", "hybrid"}
        elif isinstance(method, str):
            method = {method}

        result = {}
        if "identity" in method:
            result["identity_search"] = self.identity_search(
                precursor_mz=precursor_mz, peaks=peaks, ms1_tolerance_in_da=ms1_tolerance_in_da, ms2_tolerance_in_da=ms2_tolerance_in_da, target=target
            )
        if "open" in method:
            result["open_search"] = self.open_search(peaks=peaks, ms2_tolerance_in_da=ms2_tolerance_in_da, target=target)
        if "neutral_loss" in method:
            result["neutral_loss_search"] = self.neutral_loss_search(
                precursor_mz=precursor_mz, peaks=peaks, ms2_tolerance_in_da=ms2_tolerance_in_da, target=target
            )
        if "hybrid" in method:
            result["hybrid_search"] = self.hybrid_search(precursor_mz=precursor_mz, peaks=peaks, ms2_tolerance_in_da=ms2_tolerance_in_da, target=target)
        return result

    def build_index(
        self,
        all_spectra_list: list = None,
        max_indexed_mz: float = 1500.00005,
        precursor_ions_removal_da: float = 1.6,
        noise_threshold=0.01,
        min_ms2_difference_in_da: float = 0.05,
        max_peak_num: int = 0,
        clean_spectra: bool = True,
    ):
        """
        Set the library spectra for entropy search.

        The `all_spectra_list` must be a list of dictionaries, with each dictionary containing at least two keys: "precursor_mz" and "peaks".
        The dictionary should be in the format of {"precursor_mz": precursor_mz, "peaks": peaks, ...}, All keys in the dictionary, except "peaks,"
        will be saved as the metadata and can be accessed using the  __getitem__ function (e.g. entropy_search[0] returns the metadata for the
        first spectrum in the library).

            - The precursor_mz is the precursor m/z value of the MS/MS spectrum;

            - The peaks is an numpy array or a nested list of the MS/MS spectrum, looks like [[mz1, intensity1], [mz2, intensity2], ...].

        :param all_spectra_list:    A list of dictionaries in the format of {"precursor_mz": precursor_mz, "peaks": peaks},
                                    the spectra in the list do not need to be sorted by the precursor m/z.
                                    This function will sort the spectra by the precursor m/z and output the sorted spectra list.

        :param max_indexed_mz: The maximum m/z value that will be indexed. Default is 1500.00005.
        :param precursor_ions_removal_da:   The ions with m/z larger than precursor_mz - precursor_ions_removal_da will be removed.
                                            Default is 1.6. Set to None to not remove any ions.
        :param noise_threshold: The intensity threshold for removing the noise peaks. The peaks with intensity smaller than noise_threshold * max(intensity)
                                will be removed. Default is 0.01.
        :param min_ms2_difference_in_da:    The minimum difference between two peaks in the MS/MS spectrum. Default is 0.05.
        :param max_peak_num:    The maximum number of peaks in the MS/MS spectrum. Default is 0, which means no limit.
        :param clean_spectra:   If True, the spectra will be cleaned before indexing. Default is True. If ALL spectra in the library are pre-cleaned with the
                                function `clean_spectrum` or `clean_spectrum_for_search`, set this parameter to False. ALWAYS set this parameter to true if
                                the spectra are not pre-prepossessed with the function `clean_spectrum` or `clean_spectrum_for_search`.

        :return:    If the all_spectra_list is provided, this function will return the sorted spectra list.
        """

        # Sort the spectra by the precursor m/z.
        all_sorted_spectra_list = sorted(all_spectra_list, key=lambda x: x["precursor_mz"])

        # Clean the spectra, and collect the non-empty spectra
        all_spectra_list = []
        all_metadata_list = []
        for spec in all_sorted_spectra_list:
            # Clean the peaks
            if clean_spectra:
                spec["peaks"] = self.clean_spectrum_for_search(
                    peaks=spec["peaks"],
                    precursor_mz=spec["precursor_mz"],
                    precursor_ions_removal_da=precursor_ions_removal_da,
                    noise_threshold=noise_threshold,
                    min_ms2_difference_in_da=min_ms2_difference_in_da,
                    max_peak_num=max_peak_num,
                )
            if len(spec["peaks"]) > 0:
                all_spectra_list.append(spec)
                all_metadata_list.append(pickle.dumps(spec))

        # Extract precursor m/z array
        self.precursor_mz_array = np.array([spec["precursor_mz"] for spec in all_spectra_list], dtype=np.float32)

        # Extract metadata array
        all_metadata_len = np.array([0] + [len(metadata) for metadata in all_metadata_list], dtype=np.uint64)
        self.metadata_loc = np.cumsum(all_metadata_len).astype(np.uint64)
        self.metadata = np.frombuffer(b"".join(all_metadata_list), dtype=np.uint8)

        # Call father class to build the index.
        self.entropy_search.build_index(all_spectra_list, max_indexed_mz)
        return all_spectra_list

    def __getitem__(self, index):
        """
        Get the MS/MS metadate by the index.

        :param index:   The index of the MS/MS spectrum.
        :return:    The MS/MS spectrum in the format of {"precursor_mz": precursor_mz, "peaks": peaks}.
        """
        if self.metadata is not None:
            spectrum = pickle.loads(self.metadata[self.metadata_loc[index] : self.metadata_loc[index + 1]].tobytes())
        else:
            spectrum = {"precursor_mz": self.precursor_mz_array[index]}
        return spectrum

    def get_topn_matches(self, similarity_array, topn=3, min_similarity=0.01):
        """
        Get the topn MS/MS spectra with the highest entropy similarity.

        :param similarity_array:  The entropy similarity of the MS/MS spectra.
        :param topn:    The number of MS/MS spectra to return, if None, all the MS/MS spectra will be returned.
        :param min_similarity:  The minimum similarity of the MS/MS spectra to return, if None, all the MS/MS spectra will be returned.
        :return:    The topn MS/MS spectra with the highest entropy similarity.
        """
        if topn is None:
            topn = len(similarity_array)
        if min_similarity is None:
            min_similarity = 0.0

        # Get the topn indices
        topn_indices = np.argsort(similarity_array)[::-1][:topn]

        result = []
        for index in topn_indices:
            if similarity_array[index] < min_similarity:
                break
            item = self[index]
            item["entropy_similarity"] = similarity_array[index]
            result.append(item)
        return result

    def write(self, path_data=None):
        """
        Write the MS/MS spectral library to a file.

        :param path_data:   The path of the file to write.
        :return:    None
        """
        if path_data is None:
            path_data = self.entropy_search.path_data
        else:
            path_data = Path(path_data)

        path_data = Path(path_data)
        path_data.mkdir(parents=True, exist_ok=True)

        self.precursor_mz_array.tofile(str(path_data / "precursor_mz.npy"))
        self.metadata.tofile(str(path_data / "metadata.npy"))
        self.metadata_loc.tofile(str(path_data / "metadata_loc.npy"))

        self.entropy_search.write(path_data)

    def read(self, path_data=None):
        """
        Read the MS/MS spectral library from a file.

        :param path_data:   The path of the file to read.
        :return:    None
        """
        if path_data is None:
            path_data = self.entropy_search.path_data
        else:
            path_data = Path(path_data)

        if self.low_memory:
            self.precursor_mz_array = np.memmap(path_data / "precursor_mz.npy", dtype=np.float32, mode="r")
            self.metadata = np.memmap(path_data / "metadata.npy", dtype=np.uint8, mode="r")
            self.metadata_loc = np.memmap(path_data / "metadata_loc.npy", dtype=np.uint64, mode="r")
        else:
            self.precursor_mz_array = np.fromfile(str(path_data / "precursor_mz.npy"), dtype=np.float32)
            self.metadata = np.fromfile(str(path_data / "metadata.npy"), dtype=np.uint8)
            self.metadata_loc = np.fromfile(str(path_data / "metadata_loc.npy"), dtype=np.uint64)

        return self.entropy_search.read(path_data)

    def save_memory_for_multiprocessing(self):
        """
        Save the memory for multiprocessing. This function will move the numpy array in the index to shared memory in order to save memory.

        This function is not required when you only use one thread to search the MS/MS spectra.
        When use multiple threads, this function is also not required but highly recommended, as it avoids the memory copy and saves a lot of memory and time.

        :return:    None
        """
        self.entropy_search.save_memory_for_multiprocessing()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys

sys.path.insert(0, os.path.abspath("../../"))
sys.path.insert(0, os.path.abspath("../../ms_entropy"))

from ms_entropy import __version__

project = "MS Entropy"
copyright = "2023, Yuanyue Li"
author = "Yuanyue Li"
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", 
              "sphinx.ext.doctest", 
              "sphinx.ext.autosummary", 
              "sphinx.ext.viewcode", 
              "sphinx.ext.githubpages",
              "sphinxcontrib.email",
              "numpydoc"]

# apidoc_module_dir = "../../ms_entropy"
# apidoc_output_dir = "./api"
# apidoc_separate_modules = True
# apidoc_module_first = True
numpydoc_show_class_members = False
autoclass_content = 'both'

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



#!/usr/bin/env python3
import numpy as np
from .shared import smart_open_file


def read_one_spectrum(file_input, file_encoding='utf-8', **kwargs) -> dict:
    """
    A generator to read one spectrum from .mgf file.

    :param filename_input: The file path of input file, can be str or Path.
                            The file can be compressed with .gz, .bz2 or .zip extension.
    :param file_encoding: The encoding of input file, default is 'utf-8'.
    :return: a dict contains one spectrum information.    
    """
    fi = smart_open_file(file_input, file_encoding)

    scan_number = 1
    spectrum_start = False
    for line in fi:
        if not isinstance(line, str):
            line = line.decode()
        line = line.strip()
        if line == "BEGIN IONS":
            spectrum_start = True
            spectrum_info = {
                "_ms_level": 2,
                "_scan_number": scan_number,
                "peaks": [],
            }
            scan_number += 1
            continue
        elif line == "END IONS":
            spectrum_start = False
            # spectrum_info['peaks'] = np.array(spectrum_info['peaks']).astype(np.float32)
            yield spectrum_info
            continue

        if spectrum_start:
            if "=" in line:
                items = line.split("=", maxsplit=1)
                if len(items) == 1:
                    continue
                key, value = items
                key = key.strip().lower()
                value = value.strip()
                spectrum_info[key] = value
            else:
                items = line.split()
                if len(items)>=2:
                    spectrum_info['peaks'].append([items[0], items[1]])

    if len(spectrum_info['peaks']) > 0 or len(spectrum_info) > 3:
        yield spectrum_info

    fi.close()



from setuptools import find_packages, setup, Extension
from distutils.util import convert_path
from Cython.Build import cythonize
import numpy as np
import os

os.environ["CFLAGS"] = "-O3 -Wno-cpp -Wno-unused-function"
main_ns = {}
ver_path = convert_path("ms_entropy/version.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

common_directives = dict(
    language_level="3",
    cdivision=True,
    boundscheck=False,
    wraparound=False,
)

extensions = [
    Extension(
        "ms_entropy.spectra.entropy_cython",
        sources=[
            "ms_entropy/spectra/entropy_cython.pyx",
            "ms_entropy/spectra/CleanSpectrum.c",
            "ms_entropy/spectra/SpectralEntropy.c",
        ],
        include_dirs=[np.get_include()],
    ),
    Extension(
        "ms_entropy.entropy_search.fast_flash_entropy_search_cpython",
        sources=["ms_entropy/entropy_search/fast_flash_entropy_search_cpython.pyx"],
        include_dirs=[np.get_include()],
    ),
]

setup(
    name="ms_entropy",
    version=main_ns["__version__"],
    url="https://github.com/YuanyueLi/MSEntropy",
    platforms=["Linux", "MacOS", "Windows"],
    packages=find_packages(where=".", exclude=["tests", "docs", "examples", "manuscript", "dist", "build"]),
    package_dir={"": "."},
    python_requires=">=3.7",
    install_requires=["numpy>=1.9.13"],
    extras_require={
        "all": ["lz4>=4.3.2", "msgpack>=1.0.5", "pyteomics>=4.6"],
        "gpu": ["cupy>=12.0.0"],
    },
    ext_modules=cythonize(extensions, annotate=False, compiler_directives=common_directives),
)

# python setup.py sdist
# python setup.py build_ext --inplace


import numpy as np
cimport numpy as np

ctypedef np.float32_t float32
ctypedef np.int64_t int_64
ctypedef np.int8_t int_8
ctypedef np.uint32_t uint_32
from libc.math cimport log2


cpdef void cy_entropy_similarity_identity_search(int_64 product_mz_idx_min, int_64 product_mz_idx_max,
                                                    float32 intensity, float32[:] entropy_similarity,
                                                    const float32[:] library_peaks_intensity, const uint_32[:] library_spec_idx_array,
                                                    int_64 search_spectra_idx_min, int_64 search_spectra_idx_max) noexcept nogil:
    """
    The entropy_similarity will be modified in this function.

    Note: the intensity here should be half of the original intensity.
    """
    cdef uint_32 library_spec_idx
    cdef float32 library_peak_intensity, intensity_ab
    cdef float32 intensity_xlog2x = intensity * log2(intensity)

    for idx in range(product_mz_idx_min, product_mz_idx_max):
        library_spec_idx = library_spec_idx_array[idx]
        if  search_spectra_idx_min <= library_spec_idx and library_spec_idx < search_spectra_idx_max:
            # Match this peak
            library_peak_intensity = library_peaks_intensity[idx]
            intensity_ab = intensity + library_peak_intensity

            entropy_similarity[library_spec_idx] += \
                intensity_ab * log2(intensity_ab) - \
                intensity_xlog2x - \
                library_peak_intensity * log2(library_peak_intensity)



#!/usr/bin/env python3
import numpy as np
from .shared import smart_open_file


def read_one_spectrum(file_input, file_encoding='utf-8', **kwargs) -> dict:
    """
    A generator to read one spectrum from .msp file.

    :param filename_input: The file path of input file, can be str or Path.
                            The file can be compressed with .gz, .bz2 or .zip extension.
    :param file_encoding: The encoding of input file, default is 'utf-8'.
    :return: a dict contains one spectrum information.    
    """
    fi = smart_open_file(file_input, file_encoding)

    scan_number = 1
    spectrum_info = {
        "_ms_level": 2,
        "_scan_number": scan_number,
        "peaks": [],
    }
    peak_start = False
    peak_num = 0
    for line in fi:
        if not isinstance(line, str):
            line = line.decode()
        line = line.strip()

        # if scan_number > 10000:
        #     break

        if peak_start:
            items = line.split()
            if len(items) >= 2:
                spectrum_info['peaks'].append([items[0], items[1]])
                peak_num -= 1
            if peak_num == 0:
                # spectrum_info['peaks'] = np.array(spectrum_info['peaks']).astype(np.float32)
                yield spectrum_info
                scan_number += 1
                spectrum_info = {
                    "_ms_level": 2,
                    "_scan_number": scan_number,
                    "peaks": [],
                }
                peak_start = False
        else:
            items = line.split(":", maxsplit=1)
            if len(items) == 2:
                key, value = items
                key = key.strip().lower()
                value = value.strip()
                spectrum_info[key] = value

                if key.lower() == "num peaks":
                    peak_start = True
                    peak_num = int(value)
                    continue

    if len(spectrum_info['peaks']) > 0 or len(spectrum_info) > 3:
        yield spectrum_info
    fi.close()
